"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const eventemitter3_1 = require("eventemitter3");
const redis_utils_1 = require("./redis-utils");
class RedisQueueWorker extends eventemitter3_1.EventEmitter {
    constructor({ redis, batchSize, groupVisibilityTimeoutMs, pollingTimeoutMs, handleMessage, 
    // Redis keys
    groupStreamKey, groupSetKey, clientIndexKey, consumerGroupId, 
    // Redis script EVALSHA hashes
    callClaimTimedOutGroup, callUnlockGroup, callDeleteMessage, createPriorityMessageQueueKey }) {
        super();
        this.isRunning = false;
        this.parentRedis = redis;
        this.batchSize = batchSize;
        this.groupVisibilityTimeoutMs = groupVisibilityTimeoutMs;
        this.pollingTimeoutMs = pollingTimeoutMs;
        this.handleMessage = handleMessage;
        this.callClaimTimedOutGroup = callClaimTimedOutGroup;
        this.callUnlockGroup = callUnlockGroup;
        this.callDeleteMessage = callDeleteMessage;
        this.groupStreamKey = groupStreamKey;
        this.groupSetKey = groupSetKey;
        this.clientIndexKey = clientIndexKey;
        this.consumerGroupId = consumerGroupId;
        this._createPriorityMessageQueueKey = createPriorityMessageQueueKey;
        this.isRunning = false;
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            this.isRunning = true;
            this.redis = yield redis_utils_1.clone_redis_connection(this.parentRedis);
            const redisClientId = yield redis_utils_1.create_client_id(this.redis, this.clientIndexKey);
            this.consumerId = `consumer-${redisClientId}`;
            this._poll();
        });
    }
    stop() {
        if (!this.isRunning)
            return Promise.resolve();
        return new Promise(resolve => {
            this.isRunning = false;
            super.once('stopped', () => {
                this.redis.disconnect(false);
                resolve();
            });
        });
    }
    _poll() {
        const _super = Object.create(null, {
            emit: { get: () => super.emit }
        });
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.isRunning) {
                _super.emit.call(this, 'stopped');
                return;
            }
            const lock = yield this._lock();
            if (lock) {
                for (let batchIndex = 0; batchIndex < this.batchSize && this.isRunning; ++batchIndex) {
                    const msgData = yield this._peek(lock);
                    if (msgData) {
                        try {
                            yield this._process(lock, msgData);
                            yield this._delete(lock, msgData);
                        }
                        catch (e) {
                            console.error('Failed processing message: ', e);
                            break;
                        }
                    }
                    else {
                        break;
                    }
                }
                yield this._unlock(lock);
            }
            this._poll();
        });
    }
    _process(lock, msgData) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.handleMessage(msgData, lock);
        });
    }
    _lock() {
        return __awaiter(this, void 0, void 0, function* () {
            const claimed = yield this._claim_timed_out_lock();
            if (claimed) {
                return claimed;
            }
            const response = yield redis_utils_1.redis_call(this.redis, 'XREADGROUP', 'GROUP', this.consumerGroupId, this.consumerId, 'COUNT', 1, 'BLOCK', this.pollingTimeoutMs, 'STREAMS', this.groupStreamKey, '>');
            if (!response) {
                return null;
            }
            const [streams] = response;
            if (streams) {
                const [_, messages] = streams;
                if (messages) {
                    const [msg] = messages;
                    const [msgId, [_msgContentKey, groupId]] = msg;
                    return { groupId, messageId: msgId, queueId: this._createPriorityMessageQueueKey(groupId), consumerId: this.consumerId };
                }
            }
            return null;
        });
    }
    _claim_timed_out_lock() {
        return __awaiter(this, void 0, void 0, function* () {
            const [msgId, groupId] = yield this.callClaimTimedOutGroup([this.consumerGroupId, this.consumerId, this.groupVisibilityTimeoutMs.toString()], [this.groupStreamKey]);
            if (msgId) {
                return { groupId, messageId: msgId, queueId: this._createPriorityMessageQueueKey(groupId), consumerId: this.consumerId };
            }
            else {
                return null;
            }
        });
    }
    _unlock(lock) {
        return __awaiter(this, void 0, void 0, function* () {
            const [_ /*unlockedCount*/, remainingMessagesCount] = yield this.callUnlockGroup([this.consumerGroupId, this.consumerId, lock.messageId, lock.groupId], [this.groupStreamKey, this.groupSetKey, lock.queueId]);
            return remainingMessagesCount;
        });
    }
    _peek(lock) {
        return __awaiter(this, void 0, void 0, function* () {
            const [msgData] = yield redis_utils_1.redis_call(this.redis, 'ZRANGE', lock.queueId, 0, 0);
            return msgData;
        });
    }
    _delete(lock, messageData) {
        return __awaiter(this, void 0, void 0, function* () {
            const [_ /*removedMsgsCount*/, messageQueueLength] = yield this.callDeleteMessage([lock.groupId, messageData], [this.groupSetKey, lock.queueId]);
            return messageQueueLength;
        });
    }
}
exports.RedisQueueWorker = RedisQueueWorker;
//# sourceMappingURL=queue-worker.js.map